import { fn, symbol } from '@traversable/registry';
import { minItems } from './items.js';
import { getSchema, isRequired, property, wrapOptional } from './properties.js';
import * as Spec from './specification.js';
export { is, RAW } from './specification.js';
export { Functor, fold, unfold } from './functor.js';
export { isRequired, property, wrapOptional } from './properties.js';
export { minItems } from './items.js';
export { /* no JSON Schema representation */
SymbolSchema, VoidSchema, UndefinedSchema, BigIntSchema, NeverSchema, InlineSchema, /* data types */
UnknownSchema, AnySchema, NullSchema, BooleanSchema, IntegerSchema, NumberSchema, StringSchema, EqSchema, OptionalSchema, ArraySchema, RecordSchema, TupleSchema, UnionSchema, IntersectSchema, ObjectSchema };
/** @internal */
const Object_keys = globalThis.Object.keys;
function applyTupleOptionality(xs, {
  min,
  max
}) {
  return min === max ? xs.map(getSchema) : [...xs.slice(0, min).map(getSchema), ...xs.slice(min).map(getSchema)];
}
const NeverSchema = {
  toJsonSchema() {
    return void 0;
  }
};
const VoidSchema = {
  toJsonSchema() {
    return void 0;
  }
};
const UndefinedSchema = {
  toJsonSchema() {
    return void 0;
  }
};
const SymbolSchema = {
  toJsonSchema() {
    return void 0;
  }
};
const BigIntSchema = {
  toJsonSchema() {
    return void 0;
  }
};
function InlineSchema(_) {
  return {
    toJsonSchema: () => void 0
  };
}
/* * * * * * * * *
 *                *
 *   data types   *
 *                *
  * * * * * * * * */
const AnySchema = {
  toJsonSchema() {
    return Spec.RAW.any;
  }
};
const UnknownSchema = {
  toJsonSchema() {
    return Spec.RAW.any;
  }
};
const NullSchema = {
  toJsonSchema() {
    return Spec.RAW.null;
  }
};
const BooleanSchema = {
  toJsonSchema() {
    return Spec.RAW.boolean;
  }
};
const IntegerSchema = {
  toJsonSchema() {
    return Spec.RAW.integer;
  }
};
const NumberSchema = {
  toJsonSchema() {
    return Spec.RAW.number;
  }
};
const StringSchema = {
  toJsonSchema() {
    return Spec.RAW.string;
  }
};
function EqSchema(value) {
  return {
    toJsonSchema() {
      return {
        const: value
      };
    }
  };
}
function OptionalSchema(x) {
  function toJsonSchema() {
    return getSchema(x);
  }
  toJsonSchema[symbol.optional] = wrapOptional(x);
  return {
    toJsonSchema
  };
}
function ArraySchema(x) {
  return {
    toJsonSchema() {
      return {
        type: 'array',
        items: getSchema(x)
      };
    }
  };
}
function RecordSchema(x) {
  return {
    toJsonSchema() {
      return {
        type: 'object',
        additionalProperties: getSchema(x)
      };
    }
  };
}
function UnionSchema(xs) {
  return {
    toJsonSchema() {
      return {
        anyOf: xs.map(getSchema)
      };
    }
  };
}
function IntersectSchema(xs) {
  return {
    toJsonSchema() {
      return {
        allOf: xs.map(getSchema)
      };
    }
  };
}
function TupleSchema(xs) {
  const min = minItems(xs);
  const max = xs.length;
  return {
    toJsonSchema() {
      return {
        type: 'array',
        additionalItems: false,
        items: applyTupleOptionality(xs, {
          min,
          max
        }),
        minItems: min,
        maxItems: max
      };
    }
  };
}
function ObjectSchema(xs) {
  const required = Object_keys(xs).filter(isRequired(xs));
  return {
    toJsonSchema() {
      return {
        type: 'object',
        required,
        properties: fn.map(xs, property(required))
      };
    }
  };
}
//# sourceMappingURL=jsonSchema.js.map